/* x86_64-w64-mingw32-gcc -Wall -O0 -g3 -o map-view-of-file{.exe,.c} */

#include <io.h>
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

static const char *
last_error_message(void)
{
# define ERROR_MESSAGE_BUFFER_SIZE 4096
  int error_code = GetLastError();
  static char message[ERROR_MESSAGE_BUFFER_SIZE];
  DWORD written_bytes;

  written_bytes = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM |
                                FORMAT_MESSAGE_IGNORE_INSERTS,
                                NULL,
                                error_code,
                                MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                                message,
                                ERROR_MESSAGE_BUFFER_SIZE,
                                NULL);
  if (written_bytes >= 2) {
    if (message[written_bytes - 1] == '\n') {
      message[written_bytes - 1] = '\0';
      written_bytes--;
    }
    if (message[written_bytes - 1] == '\r') {
      message[written_bytes - 1] = '\0';
      written_bytes--;
    }
  }

  return message;

# undef ERROR_MESSAGE_BUFFER_SIZE
}

static void
show_memory_usage(int i)
{
  MEMORYSTATUSEX status;

  status.dwLength = sizeof(status);
  if (GlobalMemoryStatusEx(&status)) {
    printf("Memory status: %d\n", i);
    printf("                    Using memory size: %16ld%%\n",
           status.dwMemoryLoad);
    printf("                Total physical memory: %16I64dKiB\n",
           status.ullTotalPhys / 1024);
    printf("            Available physical memory: %16I64dKiB\n",
           status.ullAvailPhys / 1024);
    printf("               Committed memory limit: %16I64dKiB\n",
           status.ullTotalPageFile / 1024);
    printf("          Available commitable memory: %16I64dKiB\n",
           status.ullAvailPageFile / 1024);
    printf("     Total virtual address space size: %16I64dKiB\n",
           status.ullTotalVirtual / 1024);
    printf(" Available virtual address space size: %16I64dKiB\n",
           status.ullAvailVirtual / 1024);
  } else {
    printf("Failed to get memory status: %d: %s\n",
           i, last_error_message());
  }

  fflush(stdout);
}

int
main(int argc, char **argv, char **env)
{
  int i;
  DWORD map_size;
  int n_maps;
  int write_gap;

  if (argc != 4) {
    printf("Usage: %s map_size n_maps write_gap\n", argv[0]);
    return EXIT_FAILURE;
  }

  map_size = atoi(argv[1]);
  n_maps = atoi(argv[2]);
  write_gap = atoi(argv[3]);

  for (i = 0; i < n_maps; i++) {
    HANDLE file;
    HANDLE file_mapping;
    byte *map;
    char file_name[256];

    snprintf(file_name, 256, "map%d", i);
    _unlink(file_name);

    printf("Creating file: %d\n", i);
    fflush(stdout);

    show_memory_usage(i);

    file = CreateFile(file_name,
                      GENERIC_READ | GENERIC_WRITE,
                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                      NULL,
                      CREATE_NEW,
                      FILE_ATTRIBUTE_NORMAL,
                      0);
    if (!file) {
      printf("failed to CreateFile: %d: %s\n",
             i,
             last_error_message());
      return EXIT_FAILURE;
    }

    file_mapping = CreateFileMapping(file,
                                     NULL,
                                     PAGE_READWRITE,
                                     0,
                                     map_size,
                                     NULL);
    if (!file_mapping) {
      printf("failed to CreateFileMapping: %d: %s\n",
             i,
             last_error_message());
      return EXIT_FAILURE;
    }

    map = MapViewOfFile(file_mapping, FILE_MAP_WRITE, 0, 0, map_size);
    if (!map) {
      printf("failed to MapViewOfFile: %d: %s\n",
             i,
             last_error_message());
      return EXIT_FAILURE;
    }

    if (write_gap > 0) {
      DWORD j;
      for (j = 0; j < map_size; j += write_gap) {
        map[j] = 0;
      }
    }

    printf("Created file: %d\n", i);

    printf("waiting (1s)\n");
    fflush(stdout);
    Sleep(1 * 1000);
  }

  printf("waiting before shutdown (60s)\n");
  fflush(stdout);

  show_memory_usage(i);

  Sleep(60 * 1000);

  return EXIT_SUCCESS;
}
